{==============================================================================|
| Project : Notes/Delphi class library                           | 3.10        |
|==============================================================================|
| Content:                                                                     |
|==============================================================================|
| The contents of this file are subject to the Mozilla Public License Ver. 1.0 |
| (the "License"); you may not use this file except in compliance with the     |
| License. You may obtain a copy of the License at http://www.mozilla.org/MPL/ |
|                                                                              |
| Software distributed under the License is distributed on an "AS IS" basis,   |
| WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for |
| the specific language governing rights and limitations under the License.    |
|==============================================================================|
| Initial Developers of the Original Code are:                                 |
|   Daniel Lehtihet                                                            |
|   Sergey Kolchin (Russia) skolchin@yahoo.com                                 |
| All Rights Reserved.                                                         |
|   Last Modified:                                                             |
|     29.09.2002, Daniel Lehtihet                                              |
|==============================================================================|
| Contributors and Bug Corrections:                                            |
|   Fujio Kurose                                                               |
|   Noah Silva                                                                 |
|   Tibor Egressi                                                              |
|   Andreas Pape                                                               |
|   Anatoly Ivkov                                                              |
|   Winalot                                                                    |
|   Dmitry Mokrushin                                                           |
|   Olaf Hahnl                                                                 |
|   Daniel Lehtihet                                                            |
|     and others...                                                            |
|==============================================================================|
| History: see README.TXT                                                      |
|==============================================================================|
| Notes Design classes                                                         |
|                                                                              |
| Latest versions are available at                                             |
|   http://www.geocities.com/skolchin/                                         |
|==============================================================================|}
unit Class_NotesDesign;

{$INCLUDE Util_LnVersion.inc}

interface
uses SysUtils, Classes, Util_LNApi, Util_LNApiErr, Class_LotusNotes {$IFDEF D6}, Variants {$ENDIF};

type
  TNotesAgent = class;
  TNotesScriptLib = class;
  TNotesDBScript = class;

  // Exception for agent parsing errors
  EAgentError = class(ELotusNotes)
  end;

  // Database design class
  TNotesDatabaseDesign = class(TNotesDatabase)
  private
    FAgents: TStrings;
    FScriptLibs: TStrings;
    function GetAllAgentNames: TStrings;
    function GetAllScriptLibNames: TStrings;
  protected
    procedure CheckAgents;
    procedure CheckScriptLibs;
  public
    destructor Destroy; override;

    // Find design notes (generic)
    // Returns a collection containing TNotesDocumentDesign objects
    function FindDesignNotes(formula: string; notesDateTime: TDateTime; noteClass: word;
      designFlags: string): TNotesDocumentCollection;
      
    function SearchAll(formula: string; notesDateTime: TDateTime = 0; maxDocs: integer = 0): TNotesDocumentCollection;

    // Script libraries
    property ScriptLibraries: TStrings read GetAllScriptLibNames;
    function OpenLibrary(AName: string): TNotesScriptLib;

    // Agents
    property AgentNames: TStrings read GetAllAgentNames;
    function OpenAgent(AName: string): TNotesAgent;
    function OpenDBScript: TNotesDBScript;
  end;

  // Design note document
  TNotesDesignNote = class(TNotesNote)
  private
    function GetDesignFlags: string;
    function GetIsAgent: boolean;
    function GetIsForm: boolean;       //new 2003-09-28/Daniel
    function GetIsPage: boolean;       //new 2003-09-28/Daniel
    function GetIsFrameSet: boolean;   //new 2003-09-28/Daniel
    function GetIsSharedAction: boolean;    //new 2003-09-28/Daniel
    function GetIsSubForm: boolean;    //new 2003-09-28/Daniel
    function GetIsImage: boolean;      //new 2003-09-28/Daniel
    function GetIsOutLine: boolean;    //new 2003-09-28/Daniel
    function GetIsDBScript: boolean;   //new 2003-09-28/Daniel
    function GetIsScriptLib: boolean;
    procedure SetDesignFlags(const Value: string);
    procedure SetIsScriptLib(const Value: boolean);
  protected
    function GetTitle: string;
    procedure SetTitle(const Value: string);
    function GetFlag(Flag: string): boolean;
    procedure SetFlag(Flag: string; Value: boolean);
  public
    property IsScriptLib: boolean read GetIsScriptLib write SetIsScriptLib;
    property IsAgent: boolean read GetIsAgent;
    property IsForm: boolean read GetIsForm;    //new 2003-09-28/Daniel
    property IsPage: boolean read GetIsPage;    //new 2003-09-28/Daniel
    property IsFrameSet: boolean read GetIsFrameSet;    //new 2003-09-28/Daniel
    property IsSharedAction: boolean read GetIsSharedAction;    //new 2003-09-28/Daniel
    property IsSubForm: boolean read GetIsSubForm;    //new 2003-09-28/Daniel
    property IsImage: boolean read GetIsImage;       //new 2003-09-28/Daniel
    property IsOutLine: boolean read GetIsOutLine;    //new 2003-09-28/Daniel
    property IsDBScript: boolean read GetIsDBScript;  //new 2003-09-28/Daniel
    property DesignFlags: string read GetDesignFlags write SetDesignFlags;

    property Title: string read GetTitle write SetTitle;
  end;

  // Source code type
  TSourceCodeType = (scUnknown, scFormula, scLotusScript, scJava, scCompiledJava);

  // A Notes note containing program code
  TNotesCodeNote = class(TNotesDesignNote)
  private
    FFormulaHandle: Handle;
    FFormulaLength: word;
    FType: TSourceCodeType;

    //Javastuff
    FJavamainClassName: string;
    FJavaCodePath: String;
    FJavaClassNameList: String;
    FJavaClassNameArray: Array of String;

    function GetType: TSourceCodeType;
  protected
    function GetSourceCode: string; virtual; abstract;
    procedure SetSourceCode(Value: string); virtual; abstract;
    procedure SetType(const Value: TSourceCodeType); virtual;
    function GetLocalCode: string; virtual;

    procedure CompileFormula(Formula: string);
    procedure CompileLS;
    function FormatLSCode(szScript: string): string;
  public
    constructor CreateByName(notesDatabase: TNotesDatabaseDesign; aName: string);
    destructor Destroy; override;
    class function GetProgramID(notesDatabase: TNotesDatabaseDesign; aName: string): NOTEID; virtual;

    // Compile the code node
    procedure Compile; virtual;

    procedure SetJavaCode(FilePath, MainClassName: string; const JavaFileList: array of string); virtual;
      //set a list of Java files composing the code noe

    property SourceCodeType: TSourceCodeType read GetType write SetType;
    property SourceCode: string read GetSourceCode write SetSourceCode;
    property Name: string read GetTitle write SetTitle;
  end;
  TNotesCodeNoteClass = class of TNotesCodeNote;

  // Notes ScriptLib
  TNotesScriptLib = class(TNotesCodeNote)
  protected
    function GetSourceCode: string; override;
    procedure SetSourceCode(Value: string); override;
    procedure SetType(const Value: TSourceCodeType); override;
  public
    constructor CreateNew(notesDatabase: TNotesDatabase; aName: string);
  end;

  // Notes Agent
  TAgentSchedule = (SCHEDULE_EVERY_30_MINUTES,
                        SCHEDULE_EVERY_HOUR,
                        SCHEDULE_EVERY_2_HOURS,
                        SCHEDULE_EVERY_4_HOURS,
                        SCHEDULE_EVERY_8_HOURS,
                        SCHEDULE_N_HOURS_AND_MINUTES,
                        SCHEDULE_EVERY_DAY,
                        SCHEDULE_EVERY_SUNDAY,
                        SCHEDULE_EVERY_MONDAY,
                        SCHEDULE_EVERY_TUESDAY,
                        SCHEDULE_EVERY_WEDNESDAY,
                        SCHEDULE_EVERY_THURSDAY,
                        SCHEDULE_EVERY_FRIDAY,
                        SCHEDULE_EVERY_SATURDAY,
                        SCHEDULE_EVERY_MONTH,
                        SCHEDULE_NEVER,
                        SCHEDULE_MANUALLY_FROM_ACTIONS_MENU,
                        SCHEDULE_MANUALLY_FROM_AGENT_LIST,
                        SCHEDULE_BEFORE_NEW_MAIL_ARRIVES,
                        SCHEDULE_IF_NEW_MAIL_ARRIVES,
                        SCHEDULE_IF_DOCUMENTS_CREATED_OR_MODIFIED,
                        SCHEDULE_IF_DOCUMENTS_PASTED);

  TAgentAssistFlags = (AG_ASSIST_FLAG_ENABLED, AG_ASSIST_FLAG_PRIVATE, AG_ASSIST_FLAG_HIDDEN, AG_ASSIST_FLAG_DISABLED);
  TAgentActOn = (ACTON_ALL_DOCUMENTS,
                  ACTON_NEW_DOCUMENTS,
                  ACTON_MODIFIED_DOCUMENTS,
                  ACTON_SELECTED_DOCUMENTS,
                  ACTON_VIEW_DOCUMENTS,
                  ACTON_UNREAD_DOCUMENTS,
                  ACTON_UI_DOCUMENTS,
                  ACTON_NONE_DOCUMENTS);

  TNotesAgent = class(TNotesCodeNote)
  private
    FAgentHandle: LHandle;   // AgentNote
    FAgentHeader:CDACTIONHEADER;
    FAgentScript:CDACTIONLOTUSSCRIPT;
    FAgentFormula: CDACTIONFORMULA;
    FAgentJava: CDACTIONJAVAAGENT;
    FAgentInfo: ODS_ASSISTSTRUCT;
    FAgentSourceCode: String;

    function GetIsPrivate: boolean;
    procedure SetIsPrivate(Value: Boolean);
    function GetComment: string;
    procedure SetComment(const Value: string);
    function GetLastRun: TDateTime;
    function GetDocumentsProcessed: DWORD;
    function GetExitCode: DWORD;
    function GetLog: String;
    function GetAgentOwner: String;
    function GetChooseWhenEnabled: boolean;
    procedure SetChooseWhenEnabled(Value: boolean);
    function GetStoreHighLights: boolean;
    procedure SetStoreHighLights(Value: boolean);
    function GetAllowPublicAccessUsers: boolean;
    procedure SetAllowPublicAccessUsers(const Value: boolean);
    function GetIsEnabled: Boolean;
    procedure SetIsEnabled(const Value: boolean);
    function GetIsScheduled: boolean;
    procedure SetIsScheduled(Value: boolean);
    function GetRunOnWeekends: Boolean;
    procedure SetRunOnWeekends(Value: boolean);
    function GetIsHidden: Boolean;
    procedure SetIsHidden(Value: boolean);
    function GetRunAsWebUser: Boolean;
    procedure SetRunAsWebUser(Value: boolean);
    function GetAssistFlags: String;

    function GetAgentActOn: TAgentActOn;
    procedure SetRunOnServerName(const Value: String);
    function GetAssistFlag(Flag: string): boolean;
    procedure SetAssistFlag(Flag: string; Value: boolean);
    function GetAgentInfo(fieldInfo: String): String;
    function GetAgentRunInfo(fieldInfo: integer): variant;
    function GetRunOnServerName: String;
    function GetStartDate: TDateTime;
    function GetEndDate: TDateTime;
    function GetStartHour: word;
    function GetStartMinute: word;
    function GetEndHour: word;
    function GetEndMinute: word;
    function GetDayOfMonthOrWeek: word;

    procedure GetAgentInfoStruct;
    function GetSchedule: TAgentSchedule;
    procedure SetAgentType(wActionType: Word);
    procedure SetAgentLastInfo;
    procedure SetAgentTrigger(wTrigger: Word);
    procedure SetAgentAction(var AgentAction: PByte; dwActionLen: DWord);
    procedure SetAgentInfoStruct;
    procedure SetAgentRunInfo;
    procedure SetAgentVersion;

    procedure SetAssistFlags(const Value: string);
    procedure SetAgentActOn( assistSearchType: TAgentActOn);
    procedure SetSchedule( aSchedule: TAgentSchedule );
    procedure SetDayOfMonthOrWeek(const Value: word);
    procedure SetEndDate(const Value: TDateTime);
    procedure SetStartDate(const Value: TDateTime);
    function GetEndTime: TDateTime;
    function GetStartTime: TDateTime;
    procedure SetEndTime(const Value: TDateTime);
    procedure SetStartTime(const Value: TDateTime);
  protected
    procedure CreateDocument(notesDatabase: TNotesDatabase); override;
    procedure InitDocument(notesDatabase: TNotesDatabase; anID: dword); override;
    function GetSourceCode: string; override;
    procedure SetSourceCode(Value: string); override;
    procedure SetType(const Value: TSourceCodeType); override;

    procedure buildActionLS;
    procedure buildActionFormula;
    procedure buildActionJava;
  public
    constructor CreateNew(notesDatabase: TNotesDatabaseDesign; aName: string; aComment: string; shared: Boolean);
    constructor CreateByName(notesDatabase: TNotesDatabaseDesign; aName: string);
    destructor Destroy; override;
    class function GetProgramID (notesDatabase: TNotesDatabaseDesign; aName: string): NOTEID; override;

    function Run: Integer; overload;
    function Run(universalID: UNID): Integer; overload;

    {$IFDEF D4}
    procedure Save(force: boolean = False; createResponse: boolean = False; markRead: boolean = False); override;
    {$ELSE}
    procedure Save(force, createResponse, markRead: boolean); override;
    {$ENDIF}

    // Agents sheduling
    function GetAssistInfo: ODS_ASSISTSTRUCT;
    function CalcTimeAsString(chunk: integer; var hour: word; var minutes: word): string;
    procedure SetAssistInfo(AssistInfoStruct: ODS_ASSISTSTRUCT);

    procedure SetRunFromAndToDate( startDate, endDate: TDateTime );
    procedure SetStartAndEndHour( startHour,startMinutes, endHour, endMinutes: word );
    procedure SetStartTimeStartDay( startHour, startMinutes, startDay: word);
    procedure SetInterval(intervalInMinutes: word);

    // Compile the agent
    procedure SetJavaCode(FilePath, MainClassName: string; const JavaFileList: array of string); override;

    property Comment: String read GetComment write SetComment;
    property Title: String read GetTitle write SetTitle;
    property Owner: String read GetAgentOwner;
    property IsEnabled: Boolean read GetIsEnabled write SetIsEnabled;
    property AllowPublicAccessUsers: Boolean read GetAllowPublicAccessUsers write SetAllowPublicAccessUsers;
    property LastRun: TDateTime read GetLastRun;
    property DocumentsProcessed: DWORD read GetDocumentsProcessed;
    property ExitCode: DWORD read GetExitCode;
    property Log: String read GetLog;
    property IsPrivate: boolean read GetIsPrivate write SetIsPrivate;
    property IsScheduled: boolean read GetIsScheduled write SetIsScheduled;
    property IsHidden: boolean read GetIsHidden write SetIsHidden;
    property ActOn: TAgentActOn read GetAgentActOn write SetAgentActOn;
    property RunOnServerName: String read GetRunOnServerName write SetRunOnServerName;
    property RunAsWebUser: boolean read GetRunAsWebUser write SetRunAsWebUser;
    property DontRunOnWeekends: boolean read GetRunOnWeekends write SetRunOnWeekends;
    property ChooseWhenEnabled: boolean read GetChooseWhenEnabled write SetChooseWhenEnabled;
    property StoreHighLights: boolean read GetStoreHighLights write SetStoreHighLights;
    property SourceCodeType: TSourceCodeType read GetType write SetType;

    property Schedule: TAgentSchedule read GetSchedule write SetSchedule;
    property StartDate: TDateTime read GetStartDate write SetStartDate;
    property EndDate: TDateTime read GetEndDate write SetEndDate;
    property StartTime: TDateTime read GetStartTime write SetStartTime;
    property EndTime: TDateTime read GetEndTime write SetEndTime;
    property DayOfMonthOrWeek: word read GetDayOfMonthOrWeek write SetDayOfMonthOrWeek;
    property StartHour: word read GetStartHour;
    property StartMinute: word read GetStartMinute;
    property EndHour: word read GetEndHour;
    property EndMinute: word read GetEndMinute;
    property Interval: word read FAgentInfo.wInterval write SetInterval;
  end;

  // DBScript is a special kind of script lib also...
  TNotesDBScript = class(TNotesCodeNote)
        protected
            procedure SetSourceCode(Value: string); override;
            function GetSourceCode: String; override;
        public
            constructor CreateNew(notesDatabase: TNotesDatabase);
  end;

implementation
{$IFDEF D6} 
uses Variants; 
{$ENDIF} 

{ TNotesDesignNote }

function TNotesDesignNote.GetDesignFlags: string;
begin
  Result := Items[DESIGN_FLAGS].AsShortString;
end;

function TNotesDesignNote.GetFlag(Flag: string): boolean;
begin
  Result := pos(Flag, DesignFlags) <> 0;
end;

function TNotesDesignNote.GetIsAgent: boolean;
begin
  Result := (not IsScriptLib) and (
    GetFlag(DESIGN_FLAG_V4AGENT) or
    GetFlag(DESIGN_FLAG_V4PASTE_AGENT) or
    GetFlag(DESIGN_FLAG_V4PASTE_AGENT) or
    GetFlag(DESIGN_FLAG_JAVA_AGENT_WITH_SOURCE) or
    GetFlag(DESIGN_FLAG_LOTUSSCRIPT_AGENT)
  );
end;

// New 2003-09-28/Daniel
function TNotesDesignNote.GetIsImage: boolean;
begin
  Result := GetFlag(DESIGN_FLAG_IMAGE_RESOURCE);
end;

// New 2003-09-28/Daniel
function TNotesDesignNote.GetIsOutLine: boolean;
begin
  Result := GetFlag(DESIGN_FLAG_SITEMAP);
end;


// New 2003-09-28/Daniel
function TNotesDesignNote.GetIsSubForm: boolean;
begin
  Result := GetFlag(DESIGN_FLAG_SUBFORM);
end;

// New 2003-09-28/Daniel
function TNotesDesignNote.GetIsPage: boolean;
begin
  Result := GetFlag(DESIGN_FLAG_WEBPAGE);
end;

// New 2003-09-28/Daniel
function TNotesDesignNote.GetIsFrameSet: boolean;
begin
  Result := GetFlag(DESIGN_FLAG_FRAMESET);
end;

// New 2003-09-28/Daniel
function TNotesDesignNote.GetIsSharedAction: boolean;
begin
  Result := GetFlag(DESIGN_FLAG_SACTIONS);
end;

// New 2003-09-28/Daniel
function TNotesDesignNote.GetIsForm: boolean;
begin
  Result := (not IsSubForm) and
            (not IsImage) and
            (not IsPage) and
            (not IsFrameSet) and
            (not IsSharedAction);
end;

function TNotesDesignNote.GetIsScriptLib: boolean;
begin
  Result := GetFlag(DESIGN_FLAG_SCRIPTLIB);
end;
// New 2003-09-28/Daniel
function TNotesDesignNote.GetIsDBScript: boolean;
begin
  Result := GetFlag(DESIGN_FLAG_DATABASESCRIPT);
end;


function TNotesDesignNote.GetTitle: string;
begin
  Result := Items[FIELD_TITLE].AsShortString;
end;

procedure TNotesDesignNote.SetDesignFlags(const Value: string);
begin
  Items[DESIGN_FLAGS].AsShortString := Value;
end;

procedure TNotesDesignNote.SetTitle(const Value: string);
begin
  Items[FIELD_TITLE].AsString := Value;
end;

procedure TNotesDesignNote.SetFlag(Flag: string; Value: boolean);
var
  n: integer;
  s: string;
begin
  s := DesignFlags;
  if Value and (pos(Flag, s) = 0) then DesignFlags := s + Flag
  else begin
    n := pos(Flag, s);
    if n > 0 then delete(s,n,1);
    DesignFlags := s;
  end;
end;

procedure TNotesDesignNote.SetIsScriptLib(const Value: boolean);
begin
  SetFlag(DESIGN_FLAG_SCRIPTLIB, Value);
end;

//***************************************************
// TNotesDatabaseDesign
//***************************************************
destructor TNotesDatabaseDesign.Destroy;
begin
  FAgents.free;
  FScriptLibs.free;
  inherited;
end;

function TNotesDatabaseDesign.OpenAgent;
begin
  Result := TNotesAgent.CreateByName(self, aName);
end;

//***************************************************
function TNotesDatabaseDesign.OpenLibrary;
begin
  Result := TNotesScriptLib.CreateByName(self, aName);
end;

//****************************************************
function TNotesDatabaseDesign.OpenDBScript;
begin
  Result := TNotesDBScript.CreateByName(self, 'Database Script');
end;

//****************************************************
procedure TNotesDatabaseDesign.CheckScriptLibs;
var
  Coll: TNotesDocumentCollection;
  i: integer;
  Doc: TNotesDesignNote;
begin
  if FScriptLibs = nil then begin
    FScriptLibs := TStringList.create;
    Coll := FindDesignNotes('@All', 0, NOTE_CLASS_FILTER, DESIGN_FLAG_SCRIPTLIB);
    try
      for i := 0 to Coll.count-1 do begin
        Doc := Coll.Note[i] as TNotesDesignNote;
        try
          FScriptLibs.Add(Doc.Title);
        finally
          Doc.free;
        end;
      end;
    finally
      Coll.free;
    end;
  end;
end;

//****************************************************
function TNotesDatabaseDesign.SearchAll(formula: string;
  notesDateTime: TDateTime; maxDocs: integer): TNotesDocumentCollection;
begin
  Result := FindNotes (formula, notesDateTime, NOTE_CLASS_ALL, True);
end;

//****************************************************
procedure TNotesDatabaseDesign.CheckAgents;
var
  Coll: TNotesDocumentCollection;
  i: integer;
  Doc: TNotesDesignNote;
begin
  if FAgents = nil then begin
    FAgents := TStringList.create;
    Coll := FindDesignNotes('@All', 0, NOTE_CLASS_FILTER, '');
    try
      for i := 0 to Coll.count-1 do begin
        Doc := Coll.Note[i] as TNotesDesignNote;
        if Doc.IsAgent then try
          FAgents.Add(Doc.Title);
        finally
          Doc.free;
        end;
      end;
    finally
      Coll.free;
    end;
  end;
end;

//****************************************************
function TNotesDatabaseDesign.GetAllAgentNames;
begin
  CheckAgents;
  Result := FAgents;
end;

//****************************************************
function TNotesDatabaseDesign.GetAllScriptLibNames;
begin
  CheckScriptLibs;
  Result := FScriptLibs;
end;

//****************************************************
function TNotesDatabaseDesign.FindDesignNotes(formula: string;
  notesDateTime: TDateTime; noteClass: word;
  designFlags: string): TNotesDocumentCollection;
var
  s: string;
begin
  if designFlags <> '' then begin
    s := '@Contains(' + DESIGN_FLAGS + '; "' + designFlags + '")';
    if (formula = '') or (compareText(formula, '@All') = 0)
      then formula := s
      else appendStr(formula, ' & ' + s);
  end;
  Result := findNotes(formula, notesDateTime, noteClass, false);
  if Result <> nil then Result.NoteClass := TNotesDesignNote;
end;

//****************************************************
// TNotesCodeNote
//****************************************************
constructor TNotesCodeNote.CreateByName(notesDatabase: TNotesDatabaseDesign; aName: string);
begin
  inherited Create(notesDatabase, GetProgramID(notesDatabase, aName));
end;

//****************************************************
destructor TNotesCodeNote.Destroy;
begin
  SetLength(FJavaClassNameArray, 0);
  FJavaClassNameArray := nil;
  inherited Destroy;
end;

//****************************************************
class function TNotesCodeNote.GetProgramID(notesDatabase: TNotesDatabaseDesign; aName: string): NOTEID;
var
  Name2: string;
  err: dword;
begin
  Name2 := Native2Lmbcs(aName);
  err := NIFFindDesignNote(notesDatabase.Handle, PChar(Name2), NOTE_CLASS_ALLNONDATA, @Result);
  if err <> 0 then err := NIFFindPrivateDesignNote(notesDatabase.Handle, PChar(Name2),
    NOTE_CLASS_ALLNONDATA, @Result);
  CheckError(err);
end;

//****************************************************
function TNotesCodeNote.GetType: TSourceCodeType;
var
  flags: string;
begin
  if FType = scUnknown then begin
    flags := DesignFlags;
    if Pos(DESIGN_FLAG_LOTUSSCRIPT_AGENT,Flags) > 0 then FType := scLotusScript
    else if Pos(DESIGN_FLAG_JAVA_AGENT_WITH_SOURCE,Flags) > 0 then FType := scJava
    else if Pos(DESIGN_FLAG_JAVA_AGENT,Flags) > 0 then FType := scCompiledJava
    else if (Pos(DESIGN_FLAG_JAVA_AGENT_WITH_SOURCE,Flags) = 0)
     and (Pos(DESIGN_FLAG_LOTUSSCRIPT_AGENT,Flags) = 0)
     and (Pos(DESIGN_FLAG_JAVA_AGENT,Flags) = 0) then FType := scFormula
    else FType := scUnknown;
  end;
  Result := FType;
end;

//****************************************************
procedure TNotesCodeNote.SetType(const Value: TSourceCodeType);
begin
  if Value = scUnknown then raise ELotusNotes.CreateErr(-1, 'Unsupported code type');
  FType := Value;
end;

//****************************************************
procedure TNotesCodeNote.Compile;
begin
  case SourceCodeType of
    scFormula:
      CompileFormula(GetLocalCode);
    scLotusScript:
      CompileLS;
  end;
end;

//****************************************************
function TNotesCodeNote.GetLocalCode: string;
begin
  Result := GetSourceCode;
end;

//****************************************************
procedure TNotesCodeNote.CompileLS;
begin
  {$IFNDEF NOTES_R4}
  CheckError(NSFNoteLSCompile(Database.Handle, Handle, 0));
  {$ENDIF}
end;

//****************************************************
procedure TNotesCodeNote.CompileFormula(Formula: String);
var
  wdc2, wdc3, wdc4, wdc5, wdc6: word;
begin
  CheckError(NSFFormulaCompile(
             nil,                   { name of formula (none) }
             WORD(0),               { length of name }
             PChar(Formula),        { the ASCII formula }
             WORD(length(Formula)), { length of ASCII formula }
             @FFormulaHandle,       { handle to compiled formula }
             @FFormulaLength,                 { compiled formula length (don't care) }
             @wdc2,                  { return code from compile (don't care) }
             @wdc3, @wdc4, @wdc5, @wdc6)); { compile error info (don't care) }
end;

//****************************************************
procedure TNotesCodeNote.SetJavaCode(FilePath, MainClassName: string; const JavaFileList: Array of string );
var
  idx: integer;
begin
  FJavamainClassName := MainClassName;
  FJavaCodePath:= FilePath;
  SetLength(FJavaClassNameArray, System.High(JavaFileList)+1);

  // Extract from open array and build a notesified list
  for idx := 0 to System.High(JavaFileList) do begin
      FJavaClassNameList := FJavaClassNameList + JavaFileList[idx] + #0;
      // Keep track of class file names. We need them when attaching
      // the class files later on.
      FJavaClassNameArray[idx] := JavaFileList[idx];
  end;
  FJavaClassNameList := FJavaClassNameList + #0;
  SourceCodeType := scCompiledJava;
end;

//****************************************************
{$IFDEF NOTES_R4}
function TNotesCodeNote.FormatLSCode(szScript: string): string;
begin
  Result := szScript;
end;

{$ELSE}

function TNotesCodeNote.FormatLSCode(szScript: string): string;
var
  res: PChar;
  hSource, hDest, hError, hUnused: LHandle;
  err: STATUS;
begin
  Result := szScript;
  if length(szScript) = 0 then exit;

  CheckError(OSMemAlloc(0,length(szScript)+1,@hSource));
  hDest := 0;
  try
    // Copy the raw Lotus Script into the newly allocated memory space.
    res := OSLockObject(hSource);
    strPCopy(res, szScript);
    OSUnlockObject(hSource);

    // Convert the raw Lotus Script to IDE compliant format.
    err := AgentLSTextFormat(hSource,hDest,hError,0,hUnused);
    if err <> 0 then begin
      if hError = 0 then CheckError(err)
      else begin
        res := OSLockObject(hError);
        Result := string(res);
        OSUnlockObject(hError);
        OSMemFree(hError);
        raise EAgentError.CreateErr(err, Result);
      end;
    end;

    // Retreive the IDE compliant script.
    res := OSLockObject(hDest);
    Result := string(pchar(res));
    OSUnlockObject(hDest);
  finally
    OSMemFree(hSource);
    if hDest <> 0 then OSMemFree(hDest);
  end;
end;

{$ENDIF}

//****************************************************
// TNotesScriptLib
//****************************************************
const
  LIB_CODE_ITEM = '$ScriptLib';

//****************************************************
constructor TNotesScriptLib.CreateNew;
var
  wNoteClass: word;
begin
  inherited CreateNew(notesDatabase);
  wNoteClass := (NOTE_CLASS_FILTER);
  NSFNoteSetInfo(Handle, _NOTE_CLASS, @wNoteClass);
  Items['$PublicAccess'].AsString := '1';
  Name := aName;
  SetDesignFlags(
    DESIGN_FLAG_SCRIPTLIB +
    DESIGN_FLAG_HIDE_FROM_V3 +
    DESIGN_FLAG_HIDE_FROM_V4 +
    DESIGN_FLAG_QUERY_FILTER +
    DESIGN_FLAG_LOTUSSCRIPT_AGENT);
  SetType(scLotusScript);
end;

(******************************************************************************)
function CodeScanProc(Spare, ItemFlags: word;
                              Name: PChar;
                              NameLength: word;
                              Value: pointer;
                              ValueLength: dword;
                              RoutineParameter: pointer): STATUS;  stdcall;
var
  s, v: string;
  pv: pchar;
begin
  s := Lmbcs2Native(Copy(strPas(name), 1, NameLength));
  if compareText(s, LIB_CODE_ITEM) = 0 then begin
    pv := pchar(DWORD(Value)+sizeof(WORD));
    v := NotesToString(pv,ValueLength-sizeof(WORD));
    TStringList(RoutineParameter).Add(v);
  end;
  Result := NOERROR;
end;

(******************************************************************************)
function TNotesScriptLib.GetSourceCode: string;
var
  s: TStringList;
begin
  s := TStringList.Create;
  try
    CheckError(NsfItemScan(Handle, CodeScanProc, s));
    Result := s.text;
  finally
    s.free;
  end;
end;

//****************************************************
procedure TNotesScriptLib.SetType(const Value: TSourceCodeType);
begin
  // Ignore SetType because libraries are always on LotusScript
  if Value <> scLotusScript then raise ELotusNotes.CreateErr(-1, 'Unsupported code type');
end;

//****************************************************
procedure TNotesScriptLib.SetSourceCode(Value: string);
var
  formattedVal: string;
begin
  // Convert the code to a format compliant with the Notes IDE
  if SourceCodeType = scLotusScript
    then formattedVal := FormatLSCode(Native2Lmbcs(Value))
    else formattedVal := Native2Lmbcs(Value);

  Items[LIB_CODE_ITEM].IsSigned := true;
  Items[LIB_CODE_ITEM].IsSummary := false;
  Items[LIB_CODE_ITEM].AsPChar:= PChar(formattedVal);
end;

//****************************************************
// TNotesAgent
//****************************************************

//****************************************************
// Function calculates hour and minutes from assistinfo
// structure elements dwTime1 and dwTime2 where
// constants are used to set hour and minutes.
function TNotesAgent.CalcTimeAsString(chunk: integer; var hour: word; var minutes: word): string;
var
  remainder: integer;
  hourStr, minStr: string;
begin
  hour := 0;
  minutes := 0;

  hour := chunk div TICKS_IN_HOUR;
  remainder := chunk mod TICKS_IN_HOUR;
  if (remainder > 0) then begin
    remainder := TICKS_IN_HOUR div remainder;
    minutes := 60 div remainder
  end
  else begin
    minutes := remainder;
  end;
  hourStr := inttostr(hour);
  if (minutes < 10)
    then minStr := '0' + inttostr(minutes)
    else minStr := inttostr(minutes);

  Result := hourStr + ':' + minStr;
end;

procedure TNotesAgent.SetAgentActOn;
begin
  case assistSearchType of
    ACTON_ALL_DOCUMENTS:
         FAgentInfo.wSearchType := ASSISTSEARCH_TYPE_ALL;
    ACTON_NEW_DOCUMENTS:
         FAgentInfo.wSearchType := ASSISTSEARCH_TYPE_NEW;
    ACTON_MODIFIED_DOCUMENTS:
         FAgentInfo.wSearchType := ASSISTSEARCH_TYPE_MODIFIED;
    ACTON_SELECTED_DOCUMENTS:
         FAgentInfo.wSearchType := ASSISTSEARCH_TYPE_SELECTED;
    ACTON_VIEW_DOCUMENTS:
         FAgentInfo.wSearchType := ASSISTSEARCH_TYPE_VIEW;
    ACTON_UNREAD_DOCUMENTS:
         FAgentInfo.wSearchType := ASSISTSEARCH_TYPE_UNREAD;
    ACTON_UI_DOCUMENTS:
         FAgentInfo.wSearchType := ASSISTSEARCH_TYPE_UI;
    ACTON_NONE_DOCUMENTS:
         FAgentInfo.wSearchType := ASSISTSEARCH_TYPE_NONE;
  end;
end;

//****************************************************
function TNotesAgent.GetAssistInfo: ODS_ASSISTSTRUCT;
begin
  Result := FAgentInfo;
end;

//****************************************************
procedure TNotesAgent.SetAssistInfo(AssistInfoStruct: ODS_ASSISTSTRUCT);
begin
  move(AssistInfoStruct, FAgentInfo, sizeOf(ODS_ASSISTSTRUCT));
end;

//****************************************************
{procedure TNotesAgent.SetAgentSchedule(aSchedule: TAgentSchedule; assistSearchType: TAgentActOn);
begin
  SetSchedule(aSchedule);
  SetAgentActOn( assistSearchType );
end;}

function TNotesAgent.GetSchedule: TAgentSchedule;
var
  trigger: integer;
begin
  Result := SCHEDULE_NEVER;
  try
    trigger := strToInt(Items[ASSIST_TRIGGER].AsString);
  except
    exit;
  end;

  case trigger of
    ASSISTTRIGGER_TYPE_SCHEDULED: begin
      // Sheduled agents
      if FAgentInfo.wIntervalType = ASSISTINTERVAL_TYPE_MINUTES then begin
        case FAgentInfo.wInterval of
          30:   Result := SCHEDULE_EVERY_30_MINUTES;
          60:   Result := SCHEDULE_EVERY_HOUR;
          120:  Result := SCHEDULE_EVERY_2_HOURS;
          240:  Result := SCHEDULE_EVERY_4_HOURS;
          480:  Result := SCHEDULE_EVERY_8_HOURS;
          else Result := SCHEDULE_N_HOURS_AND_MINUTES;
        end;
      end
      else if FAgentInfo.wInterval = 1 then
        case FAgentInfo.wIntervalType of
          ASSISTINTERVAL_TYPE_DAYS:
            Result := SCHEDULE_EVERY_DAY;
          ASSISTINTERVAL_TYPE_MONTH:
            Result := SCHEDULE_EVERY_MONTH;
          ASSISTINTERVAL_TYPE_WEEK:
            Result := TAgentSchedule(ord(SCHEDULE_EVERY_SUNDAY) + FAgentInfo.dwTime2 - 1);
        end;
    end;

    ASSISTTRIGGER_TYPE_SYNCHNEWMAIL:
      Result := SCHEDULE_BEFORE_NEW_MAIL_ARRIVES;

    ASSISTTRIGGER_TYPE_NEWMAIL:
      Result := SCHEDULE_IF_NEW_MAIL_ARRIVES;

    ASSISTTRIGGER_TYPE_DOCUPDATE:
      Result := SCHEDULE_IF_DOCUMENTS_CREATED_OR_MODIFIED;

    ASSISTTRIGGER_TYPE_PASTED:
      Result := SCHEDULE_IF_DOCUMENTS_PASTED;

    ASSISTTRIGGER_TYPE_MANUAL: begin
      if IsHidden
        then Result := SCHEDULE_MANUALLY_FROM_AGENT_LIST
        else Result := SCHEDULE_MANUALLY_FROM_ACTIONS_MENU;
    end;
  end;
end;

//****************************************************
procedure TNotesAgent.SetSchedule(aSchedule: TAgentSchedule);
begin
  // Set the Flag to determine that the agent is scheduled
  SetFlag(DESIGN_FLAG_V4BACKGROUND_MACRO, true);

  // And default values
  Items[ASSIST_TRIGGER].AsString := IntToStr(ASSISTTRIGGER_TYPE_SCHEDULED);
  FAgentInfo.wTriggerType := ASSISTTRIGGER_TYPE_SCHEDULED;
  FAgentInfo.wIntervalType := ASSISTINTERVAL_TYPE_MINUTES;
  FAgentInfo.dwTime1 := 1*TICKS_IN_HOUR;  // 01.00 AM
  FAgentInfo.dwTime2 := 23*TICKS_IN_HOUR; // 11.00 PM

   case aSchedule of
     SCHEDULE_EVERY_30_MINUTES:
            FAgentInfo.wInterval :=  30;
     SCHEDULE_EVERY_HOUR:
           FAgentInfo.wInterval :=  60;
     SCHEDULE_EVERY_2_HOURS:
           FAgentInfo.wInterval :=  120;
     SCHEDULE_EVERY_4_HOURS:
           FAgentInfo.wInterval :=  240;
     SCHEDULE_EVERY_8_HOURS:
           FAgentInfo.wInterval :=  480;
     SCHEDULE_EVERY_DAY:
     begin
           FAgentInfo.wIntervalType := ASSISTINTERVAL_TYPE_DAYS;
           FAgentInfo.wInterval :=  1;     // once every day
           FAgentInfo.dwTime1 := 12*TICKS_IN_HOUR;
           FAgentInfo.dwTime2 := 0; // ignored
     end;
     SCHEDULE_EVERY_MONTH:
     begin
           FAgentInfo.wIntervalType := ASSISTINTERVAL_TYPE_MONTH;
           FAgentInfo.wInterval :=  1;     // once every month
           FAgentInfo.dwTime1 := 12*TICKS_IN_HOUR;
           FAgentInfo.dwTime2 := 1; // default to first day of month
     end;
     SCHEDULE_EVERY_SUNDAY:
     begin
           FAgentInfo.wIntervalType := ASSISTINTERVAL_TYPE_WEEK;
           FAgentInfo.wInterval :=  1;     // once every week
           FAgentInfo.dwTime1 := 12*TICKS_IN_HOUR;
           FAgentInfo.dwTime2 := 1; // sun=1, mon=2....etc
     end;
     SCHEDULE_EVERY_MONDAY:
     begin
           FAgentInfo.wIntervalType := ASSISTINTERVAL_TYPE_WEEK;
           FAgentInfo.wInterval :=  1;     // once every week
           FAgentInfo.dwTime1 := 12*TICKS_IN_HOUR;
           FAgentInfo.dwTime2 := 2; // sun=1, mon=2....etc
     end;
     SCHEDULE_EVERY_TUESDAY:
     begin
           FAgentInfo.wIntervalType := ASSISTINTERVAL_TYPE_WEEK;
           FAgentInfo.wInterval :=  1;     // once every week
           FAgentInfo.dwTime1 := 12*TICKS_IN_HOUR;
           FAgentInfo.dwTime2 := 3; // sun=1, mon=2....etc
     end;
     SCHEDULE_EVERY_WEDNESDAY:
      begin
           FAgentInfo.wIntervalType := ASSISTINTERVAL_TYPE_WEEK;
           FAgentInfo.wInterval :=  1;     // once every week
           FAgentInfo.dwTime1 := 12*TICKS_IN_HOUR;
           FAgentInfo.dwTime2 := 4; // sun=1, mon=2....etc
      end;
     SCHEDULE_EVERY_THURSDAY:
      begin
           FAgentInfo.wIntervalType := ASSISTINTERVAL_TYPE_WEEK;
           FAgentInfo.wInterval :=  1;     // once every week
           FAgentInfo.dwTime1 := 12*TICKS_IN_HOUR;
           FAgentInfo.dwTime2 := 5; // sun=1, mon=2....etc
      end;
     SCHEDULE_EVERY_FRIDAY:
      begin
           FAgentInfo.wIntervalType := ASSISTINTERVAL_TYPE_WEEK;
           FAgentInfo.wInterval :=  1;     // once every week
           FAgentInfo.dwTime1 := 12*TICKS_IN_HOUR;
           FAgentInfo.dwTime2 := 6; // sun=1, mon=2....etc
      end;
     SCHEDULE_EVERY_SATURDAY:
      begin
           FAgentInfo.wIntervalType := ASSISTINTERVAL_TYPE_WEEK;
           FAgentInfo.wInterval :=  1;     // once every week
           FAgentInfo.dwTime1 := 12*TICKS_IN_HOUR;
           FAgentInfo.dwTime2 := 7; // sun=1, mon=2....etc
      end;
      SCHEDULE_BEFORE_NEW_MAIL_ARRIVES:
      begin
           Items[ASSIST_TRIGGER].AsString := IntToStr(ASSISTTRIGGER_TYPE_SYNCHNEWMAIL);
           FAgentInfo.wTriggerType := ASSISTTRIGGER_TYPE_SYNCHNEWMAIL;
           FAgentInfo.wSearchType := ASSISTSEARCH_TYPE_NONE;
           FAgentInfo.wIntervalType := ASSISTINTERVAL_TYPE_NONE;
      end;
      SCHEDULE_IF_NEW_MAIL_ARRIVES:
      begin
           Items[ASSIST_TRIGGER].AsString := IntToStr(ASSISTTRIGGER_TYPE_NEWMAIL);
           FAgentInfo.wTriggerType := ASSISTTRIGGER_TYPE_NEWMAIL;
           FAgentInfo.wSearchType := ASSISTSEARCH_TYPE_NONE;
           FAgentInfo.wIntervalType := ASSISTINTERVAL_TYPE_NONE;
      end;
      SCHEDULE_IF_DOCUMENTS_CREATED_OR_MODIFIED:
      begin
           Items[ASSIST_TRIGGER].AsString := IntToStr(ASSISTTRIGGER_TYPE_DOCUPDATE);
           FAgentInfo.wTriggerType := ASSISTTRIGGER_TYPE_DOCUPDATE;
           FAgentInfo.wSearchType := ASSISTSEARCH_TYPE_NONE;
           FAgentInfo.wIntervalType := ASSISTINTERVAL_TYPE_NONE;
      end;
      SCHEDULE_IF_DOCUMENTS_PASTED:
      begin
           Items[ASSIST_TRIGGER].AsString := IntToStr(ASSISTTRIGGER_TYPE_PASTED);
           FAgentInfo.wTriggerType := ASSISTTRIGGER_TYPE_PASTED;
           FAgentInfo.wSearchType := ASSISTSEARCH_TYPE_NONE;
           FAgentInfo.wIntervalType := ASSISTINTERVAL_TYPE_NONE;
      end;
      SCHEDULE_MANUALLY_FROM_ACTIONS_MENU:
      begin
           Items[ASSIST_TRIGGER].AsString := IntToStr(ASSISTTRIGGER_TYPE_MANUAL);
           FAgentInfo.wTriggerType := ASSISTTRIGGER_TYPE_MANUAL;
           FAgentInfo.wIntervalType := ASSISTINTERVAL_TYPE_NONE;
      end;
      SCHEDULE_MANUALLY_FROM_AGENT_LIST:
      begin
           Items[ASSIST_TRIGGER].AsString := IntToStr(ASSISTTRIGGER_TYPE_MANUAL);
           FAgentInfo.wTriggerType := ASSISTTRIGGER_TYPE_MANUAL;
           FAgentInfo.wIntervalType := ASSISTINTERVAL_TYPE_NONE;
           SetIsHidden( true );
      end;
  end; // Schedule portion (Every_X_Z)
end;

//****************************************************
function TNotesAgent.GetAgentActOn;
begin
        case FAgentInfo.wSearchType of
            ASSISTSEARCH_TYPE_ALL: result := ACTON_ALL_DOCUMENTS;
            ASSISTSEARCH_TYPE_NEW: result := ACTON_NEW_DOCUMENTS;
            ASSISTSEARCH_TYPE_MODIFIED: result := ACTON_MODIFIED_DOCUMENTS;
            ASSISTSEARCH_TYPE_SELECTED: result := ACTON_SELECTED_DOCUMENTS;
            ASSISTSEARCH_TYPE_VIEW: result := ACTON_VIEW_DOCUMENTS;
            ASSISTSEARCH_TYPE_UNREAD: result := ACTON_UNREAD_DOCUMENTS;
        else
            result := ACTON_NONE_DOCUMENTS;
        end;
end;

//****************************************************
procedure TNotesAgent.SetRunFromAndToDate( startDate, endDate: TDateTime );
begin
  FAgentInfo.StartTime := DateTimeToNotes( startDate );
  FAgentInfo.Endtime := DateTimeToNotes( endDate );
end;

//****************************************************
procedure TNotesAgent.SetStartTimeStartDay( startHour, startMinutes, startDay: word );
begin
  FAgentInfo.dwTime1 := startHour*TICKS_IN_HOUR +
                        startMinutes*TICKS_IN_MINUTE;
  FAgentInfo.dwTime2 := startDay;
end;

//****************************************************
procedure TNotesAgent.SetStartAndEndHour( startHour, startMinutes, endHour, endMinutes: word );
begin
  FAgentInfo.dwTime1 := startHour*TICKS_IN_HOUR +
                        startMinutes*TICKS_IN_MINUTE;
  FAgentInfo.dwTime2 := endHour*TICKS_IN_HOUR +
                        endMinutes*TICKS_IN_MINUTE;
end;

//****************************************************
procedure TNotesAgent.SetInterval(intervalInMinutes: word);
begin
  FAgentInfo.wInterval := intervalInMinutes;
end;

//****************************************************
function TNotesAgent.GetStartDate: TDateTime;
begin
  result := NotesToDateTime(FAgentInfo.StartTime);
end;

//****************************************************
function TNotesAgent.GetEndDate: TDateTime;
begin
  result := NotesToDateTime(FAgentInfo.EndTime);
end;

//****************************************************
function TNotesAgent.GetStartHour: word;
var
  hour, minutes: word;
begin
  if (FAgentInfo.dwTime1 <= 0) then result := 0
  else begin
    calcTimeAsString(FAgentInfo.dwTime1, hour, minutes);
    result := hour
  end;
end;

//****************************************************
function TNotesAgent.GetStartMinute: word;
var
  hour, minutes: word;
begin
  if (FAgentInfo.dwTime1 <= 0) then result := 0
  else begin
    calcTimeAsString(FAgentInfo.dwTime1, hour, minutes);
    result := minutes
  end;
end;

//****************************************************
function TNotesAgent.GetEndHour: word;
var
  hour, minutes: word;
begin
  Result := 0;
  if (FAgentInfo.dwTime2 > 0) then
    case FAgentInfo.wIntervalType of
      ASSISTINTERVAL_TYPE_MINUTES,
      ASSISTINTERVAL_TYPE_DAYS: begin
          calcTimeAsString(FAgentInfo.dwTime2, hour, minutes);
          result := hour
      end;
      ASSISTINTERVAL_TYPE_WEEK,
      ASSISTINTERVAL_TYPE_MONTH: Result := word(FAgentInfo.dwTime2);
    end;
end;

//****************************************************
function TNotesAgent.GetEndMinute: word;
var
  hour, minutes: word;
begin
  Result := 0;
  if (FAgentInfo.dwTime2 > 0) then
    case FAgentInfo.wIntervalType of
      ASSISTINTERVAL_TYPE_MINUTES,
      ASSISTINTERVAL_TYPE_DAYS: begin
          calcTimeAsString(FAgentInfo.dwTime2, hour, minutes);
          result := minutes
      end;
    end;
end;

//****************************************************
function TNotesAgent.GetDayOfMonthOrWeek: word;
begin
  result := GetEndHour;
end;
//****************************************************

{function TNotesAgent.GetAgentSchedule;
begin
  //Result := Items[ASSIST_TRIGGER].AsString;
end;
}
//****************************************************

procedure TNotesAgent.SetSourceCode(Value: string);
begin
   // Convert the code to a format compliant with the Notes IDE
   case SourceCodeType of
        scLotusScript: begin
                FAgentSourceCode := FormatLSCode(PChar(Native2Lmbcs(Value)) );
                buildActionLS;
        end;
        scFormula: begin
                FAgentSourceCode := Native2Lmbcs( Value );
                buildActionFormula;
        end;
   end;
end;

//****************************************************
procedure TNotesAgent.SetJavaCode(FilePath, MainClassName: string; const JavaFileList: array of string);
begin
  inherited;
  buildActionJava;
end;

constructor TNotesAgent.CreateNew;
var
  wNoteClass:word;
begin
  inherited CreateNew(notesDatabase);

  // First clear the $AssistInfo structure/record
  fillchar (FAgentInfo,  sizeof(FAgentInfo), 0);

  // Clear existing flags before setting them
  //SetDesignFlags('');

  SetFlag(DESIGN_FLAG_PRIVATE_IN_DB, true);
  Items[FIELD_TITLE].IsSigned := true;
  Items[DESIGN_FLAGS].IsSummary := false;
  Items[DESIGNER_VERSION_ITEM].AsString := '5';

  if (shared = false) then begin
    wNoteClass := (NOTE_CLASS_FILTER or NOTE_CLASS_PRIVATE);
    DesignFlags := DESIGN_FLAG_PRIVATE_IN_DB;
  end
  else begin
    wNoteClass := NOTE_CLASS_FILTER;
  end;

  NSFNoteSetInfo(Handle, _NOTE_CLASS, @wNoteClass);
  Comment := aComment;
  SetAgentLastInfo;  // Create an empty container for log data

  // Set init agent version
  SetAgentVersion;
  Title := aName;
end;

//****************************************************
constructor TNotesAgent.CreateByName(notesDatabase: TNotesDatabaseDesign; aName: string);
begin
  inherited CreateByName(notesDatabase, aName);

  // Load agent data
  GetAgentInfoStruct;
  FAgentSourceCode := GetSourceCode;
end;

//****************************************************
procedure TNotesAgent.CreateDocument(notesDatabase: TNotesDatabase);
begin
  inherited;
end;

//****************************************************
procedure TNotesAgent.InitDocument(notesDatabase: TNotesDatabase; anID: dword);
begin
  inherited;
  CheckError(AgentOpen(notesDatabase.Handle, anID, FAgentHandle));
end;

//****************************************************
destructor TNotesAgent.Destroy;
begin
  if FAgentHandle <> 0 then CheckError(AgentClose(FAgentHandle));
  inherited Destroy;
end;

//****************************************************
function TNotesAgent.Run: Integer;
var
  FAgentCtx: LHandle;
begin
  Result := 0;
  CheckError(AgentCreateRunContext(FAgentHandle, '', 0, FAgentCtx));
  try
    CheckError(AgentSetDocumentContext(FAgentCtx, Handle));
    CheckError(AgentRun(FAgentHandle, FAgentCtx, 0, 0));
  finally
    CheckError(AgentDestroyRunContext(FAgentCtx));
  end;
end;

//****************************************************
function TNotesAgent.Run(universalID: UNID): Integer;
var
  FAgentCtx: LHandle;
  doc: TNotesDocument;
begin
  Result := 0;
  CheckError(AgentCreateRunContext(FAgentHandle, '', 0, FAgentCtx));
  doc := nil;
  try
    doc := TNotesDocument.CreateByUNID(Database, universalID);
    CheckError(AgentSetDocumentContext(FAgentCtx, doc.Handle));
    CheckError(AgentRun(FAgentHandle, FAgentCtx, 0, 0));
  finally
    doc.free;
    CheckError(AgentDestroyRunContext(FAgentCtx));
  end;
end;

//****************************************************
{procedure TNotesAgent.Compile;
begin
  case SourceCodeType of
  scLotusScript:
      CompileLS;
  scFormula:
    CompileFormula( FAgentSourceCode );
  end;
end;}

//****************************************************
function TNotesAgent.GetAgentInfo;
begin
  if not IsItemExists(fieldInfo)
    then Result := ''
    else Result := Items[fieldInfo].AsString;
end;

//****************************************************
function TNotesAgent.GetAgentRunInfo;
var
  dwItemSize: dword;
  wDataType: word;
  bidRunInfo: BLOCKID;
  pBuf: pchar;
  wCounter: Word;
  objRunInfo: OBJECT_DESCRIPTOR;
  dwObjectSize: dword;
  wClass,wPrivs, dwOffset: Word;
  hBuffer: LHandle;
  RunHeader: ODS_ASSISTRUNOBJECTHEADER;
  RunInfo: ODS_ASSISTRUNINFO;
  RunEntry: array[0..5] of ODS_ASSISTRUNOBJECTENTRY;
begin
  Result := Unassigned;
  FillChar(RunInfo, sizeof(RunInfo), 0);
  FillChar(RunHeader, sizeof(RunHeader), 0);

  CheckError(NSFItemInfo(Handle, ASSIST_RUNINFO_ITEM,
              length (ASSIST_RUNINFO_ITEM), nil,
              @wDataType, @bidRunInfo, @dwItemSize));
  pBuf := OsLockBlock(bidRunInfo);
  try
    // Read in the OBJECT_DESCRIPTOR ODS record and ensure that there are data objects
    pBuf := pbuf + sizeOf(word);
    ODSReadMemory( @pBuf, _OBJECT_DESCRIPTOR, @objRunInfo, 1 );
  finally
    OSUnlockBlock(bidRunInfo);
  end;

  CheckError(NSFDbGetObjectSize (Database.Handle, objRunInfo.RRV, objRunInfo.ObjectType,
                                 dwObjectSize, wClass, wPrivs ));
  if dwObjectSize = 0 then raise EAgentError.Create('Invalid run data object item');

{*
* ODS_ASSISTRUNOBJECTHEADER - contains the number of run data object entries.
* By default, there are at least 5 objects reserved for run data.  Of
* importance, the first of these always contains the RunInfo data.   The
* third of these always contains an agent execution log.
*}
  dwOffset := 0;
  CheckError(NSFDbReadObject(Database.Handle, objRunInfo.RRV, dwOffset,
                      ODSLength(_ODS_ASSISTRUNOBJECTHEADER), hBuffer));

  pBuf := OSLockObject(hBuffer);
  try
    ODSReadMemory( @pBuf, _ODS_ASSISTRUNOBJECTHEADER, @RunHeader, 1 );
  finally
    OSUnlockObject(hBuffer);
    OSMemFree(hBuffer);
  end;
  if RunHeader.wEntries > 0 then begin
    {*
    * ODS_ASSISTRUNOBJECTENTRY - determines which entry has run data.
    * If the entry length of both the run info data and log objects are greater
    * than zero, than the agent has been executed at least once and run data
    * exists.  *}
    dwOffset := dwOffset + ODSLength(_ODS_ASSISTRUNOBJECTHEADER);
    CheckError(NSFDbReadObject(Database.Handle,objRunInfo.RRV, dwOffset,
                    ODSLength(_ODS_ASSISTRUNOBJECTENTRY)*RunHeader.wEntries, hBuffer));

    {* read the first entry and make sure that a RunInfo data object exists. *}
    pBuf := OSLockObject(hBuffer);
    try
      ODSReadMemory( @pBuf, _ODS_ASSISTRUNOBJECTENTRY, @RunEntry[0], 1 );
      if RunEntry[0].dwLength <> 0 then begin

        {* read the third entry and make sure that an execution log object exists.
        * Note - we check this only because the Notes UI creates an empty RunInfo
        * data object when an agent is executed, but only creates the log object
        * if executed.
        *}
        ODSReadMemory( @pBuf, _ODS_ASSISTRUNOBJECTENTRY, @RunEntry[1], 1 );
        ODSReadMemory( @pBuf, _ODS_ASSISTRUNOBJECTENTRY, @RunEntry[2], 1 );

        {* and skip over the remaining reserved run data objects *}
        for wCounter := 3 to RunHeader.wEntries do
          ODSReadMemory( @pBuf, _ODS_ASSISTRUNOBJECTENTRY, @RunEntry[wCounter], 1 );
        dwOffset := dwOffset + ODSLength(_ODS_ASSISTRUNOBJECTENTRY)*RunHeader.wEntries;
      end;
    finally
      OSUnlockObject(hBuffer);
      OSMemFree(hBuffer);
    end;
  end;

{*
* ODS_ASSISTRUNINFO - contain the Run Information.
* It is always the first run data object aftet offest.
*}
  if dwOffset + ODSLength(_ODS_ASSISTRUNINFO) <= dwObjectSize then begin
    CheckError(NSFDbReadObject(Database.Handle, objRunInfo.RRV, dwOffset,
                      ODSLength(_ODS_ASSISTRUNINFO), hBuffer));

    pBuf := OSLockObject(hBuffer);
    try
      ODSReadMemory( @pBuf, _ODS_ASSISTRUNINFO, @RunInfo, 1 );
    finally
      OSUnlockObject(hBuffer);
      OSMemFree(hBuffer);
    end;
  end;

  case fieldInfo of
    AGENT_LAST_RUN:
      Result := NotesToDateTime(RunInfo.LastRun);
    AGENT_DOCUMENTS_PROCESSED:
      Result := RunInfo.dwProcessed;
    AGENT_EXIT_CODE:
      Result := RunInfo.dwExitCode;
    AGENT_LOG: begin
      if (RunEntry[0].dwLength <> 0) and (RunEntry[2].dwLength <> 0) then begin
        // skip the next one, display the execution log, and dump the rest
        dwOffset := dwOffset + RunEntry[0].dwLength;
        for wCounter := 1 to RunHeader.wEntries - 1 do begin
          if (RunEntry[wCounter].dwLength <> 0) then begin
            CheckError(NSFDbReadObject(Database.Handle, objRunInfo.RRV, dwOffset,
              RunEntry[wCounter].dwLength, hBuffer));

            pBuf := OSLockObject(hBuffer);
            try
              if (wCounter = 2) then begin
                pBuf[RunEntry[wCounter].dwLength] := '0';
                Result := String(pBuf);
                break;
              end;
              dwOffset := dwOffset + RunEntry[wCounter].dwLength;
            finally
              OSUnlockObject(hBuffer);
              OSMemFree(hBuffer);
            end;
          end;
        end
      end;
    end;
  end;
end;

//****************************************************
function TNotesAgent.GetComment;
begin
  Result := GetAgentInfo('$Comment');
end;

//****************************************************
function TNotesAgent.GetAgentOwner;
begin
  Result := GetAgentInfo('$UpdatedBy');
end;

//****************************************************
function TNotesAgent.GetIsEnabled;
begin
  Result := AgentIsEnabled(FAgentHandle);
end;

//****************************************************
procedure TNotesAgent.SetIsEnabled(const Value: boolean);
begin
        SetAssistFlag( ASSIST_FLAG_ENABLED, Value);
end;
//****************************************************
procedure TNotesAgent.SetComment(const Value: string);
begin
  Items['$Comment'].AsString := Value;
end;

//****************************************************
function TNotesAgent.GetAllowPublicAccessUsers: boolean;
begin
     Result := (IsItemExists(FIELD_PUBLICACCESS) and
        (Items[FIELD_PUBLICACCESS].AsString = '1'));
end;
//****************************************************
procedure TNotesAgent.SetAllowPublicAccessUsers(const Value: boolean);
begin
        if (Value) then Items[FIELD_PUBLICACCESS].AsString := '1'
        else
           if IsItemExists(FIELD_PUBLICACCESS) then DeleteItem(FIELD_PUBLICACCESS);
end;

//****************************************************
function TNotesAgent.GetLastRun;
var
  v: variant;
begin
  v := GetAgentRunInfo(AGENT_LAST_RUN);
  if VarIsEmpty(v) or VarIsNull(v)
    then Result := 0
    else Result := v;
end;

//****************************************************
function TNotesAgent.GetDocumentsProcessed;
var
  v: variant;
begin
  v := GetAgentRunInfo(AGENT_DOCUMENTS_PROCESSED);
  if VarIsEmpty(v) or VarIsNull(v)
    then Result := 0
    else Result := v;
end;

//****************************************************
function TNotesAgent.GetExitCode;
var
  v: variant;
begin
  v := GetAgentRunInfo(AGENT_EXIT_CODE);
  if VarIsEmpty(v) or VarIsNull(v)
    then Result := 0
    else Result := v;
end;

//****************************************************
function TNotesAgent.GetLog;
var
  v: variant;
begin
  v := GetAgentRunInfo(AGENT_LOG);
  if VarIsEmpty(v) or VarIsNull(v)
    then Result := ''
    else Result := v;
end;

//****************************************************
class function TNotesAgent.GetProgramID;
var
  Name2: string;
  err: dword;
  n: integer;
begin
  Name2 := Native2Lmbcs(aName);
  err := NIFFindDesignNote (notesDatabase.Handle, PChar(Name2), NOTE_CLASS_FILTER, @Result);
  if err <> 0 then err := NIFFindPrivateDesignNote (notesDatabase.Handle, PChar(Name2),
    NOTE_CLASS_FILTER, @Result);
  if err <> 0 then begin
    // Do the same for Agent alias (if present)
    n := LastDelimiter('|',aName);
    if n > 0 then begin
      Name2 := Native2Lmbcs(copy(aName,n+1,length(aName)-n));
      err := NIFFindDesignNote (notesDatabase.Handle, PChar(Name2), NOTE_CLASS_FILTER, @Result);
      if err <> 0 then err := NIFFindPrivateDesignNote (notesDatabase.Handle, PChar(Name2),
        NOTE_CLASS_FILTER, @Result);
    end;
  end;
  CheckError(err);
end;

//****************************************************
function TNotesAgent.GetSourceCode: String;
var
  dwItemSize: dword;
  wDataType, wFormulaTextLength: word;
  bidRunInfo: BLOCKID;
  pBuf, pcFormulaText: pchar;
  hFormulaText: LHandle;
  agHeader: CDACTIONHEADER;
  agInfo: CDACTIONLOTUSSCRIPT;
  maInfo: CDACTIONFORMULA;
  num: NUMBER;
begin
  CheckError(NSFItemInfo(Handle, ASSIST_ACTION,
              length (ASSIST_ACTION), NIL,
              @wDataType, @bidRunInfo, @dwItemSize));

  // Read $AssistType item to determine type of agent
  NSFItemGetNumber(Handle, ASSIST_TYPE_ITEM, @num);

  { assign pointer to item, and skip over TYPE_OBJECT word }
  pBuf := OsLockBlock(bidRunInfo);
  try
    pBuf := pbuf + sizeOf(word);
    ODSReadMemory( @pBuf, _CDACTIONHEADER, @agHeader, 1 ); // ACTION HEADER

    if (num = SIG_ACTION_LOTUSSCRIPT) then
    begin
      ODSReadMemory( @pBuf, _CDACTIONLOTUSSCRIPT, @agInfo, 1 );
      setLength(Result, agInfo.dwScriptLen+1);
      StrLCopy(pchar(Result), pBuf, agInfo.dwScriptLen);
      Result[agInfo.dwScriptLen+1] := #0;
      Result := string(pchar(Result));
    end
    else if (num = SIG_ACTION_FORMULA) or (num = SIG_ACTION_FORMULAONLY) then begin
      ODSReadMemory( @pBuf, _CDACTIONFORMULA, @maInfo, 1 );
      NSFFormulaDecompile( pBuf, FALSE, @hFormulaText, @wFormulaTextLength);
      pcFormulaText := OSLockObject(hFormulaText);
      try
        setLength( Result, wFormulaTextLength+1);
        StrLCopy(pchar(Result), pcFormulaText, wFormulaTextLength);
        Result[wFormulaTextLength+1] := #0;
        Result := string(pchar(Result));
      finally
        OSUnlockObject(hFormulaText);
      end;
    end
    else Result := '';
  finally
    OSUnlockBlock(bidRunInfo);
  end;
end;

function TNotesAgent.GetIsScheduled: boolean;
begin
  Result := GetFlag(DESIGN_FLAG_V4BACKGROUND_MACRO);
end;

//****************************************************
procedure TNotesAgent.SetIsScheduled(Value: boolean);
begin
  SetFlag(DESIGN_FLAG_V4BACKGROUND_MACRO,Value);
end;

//****************************************************
function TNotesAgent.GetIsPrivate: boolean;
begin
  Result := GetFlag(DESIGN_FLAG_PRIVATE_IN_DB);;
end;

//****************************************************
procedure TNotesAgent.SetIsPrivate(Value: Boolean);
begin
  SetFlag(DESIGN_FLAG_PRIVATE_IN_DB, Value);
  SetAssistFlag(ASSIST_FLAG_PRIVATE, Value);
end;

//****************************************************
function TNotesAgent.GetRunOnServerName: String;
begin
  Result := Items[AGENT_MACHINE_NAME].AsString;
end;

//****************************************************
procedure TNotesAgent.SetRunOnServerName(const Value: String);
var
  aName: string;
begin
  aName := TNotesName.TranslateName(Value, true, '');
  Items[AGENT_MACHINE_NAME].IsSigned := true;
  Items[AGENT_MACHINE_NAME].AsString := aName;
end;

//****************************************************
function TNotesAgent.GetRunAsWebUser: boolean;
begin
  Result := GetFlag(DESIGN_FLAG_AGENT_RUNASWEBUSER);
end;

//****************************************************
procedure TNotesAgent.SetRunAsWebUser(Value: boolean);
begin
  SetFlag(DESIGN_FLAG_AGENT_RUNASWEBUSER, Value);
end;

//****************************************************
function TNotesAgent.GetIsHidden: boolean;
begin
  result := ASSISTODS_FLAG_HIDDEN = (FAgentInfo.dwFlags and integer(ASSISTODS_FLAG_HIDDEN));
end;

//****************************************************
procedure TNotesAgent.SetIsHidden(Value: boolean);
begin
  SetAssistFlag(ASSIST_FLAG_HIDDEN, Value);

  if (Value = true) then
    FAgentInfo.dwFlags := FAgentInfo.dwFlags xor ASSISTODS_FLAG_HIDDEN
  else
    FAgentInfo.dwFlags := FAgentInfo.dwFlags or ASSISTODS_FLAG_HIDDEN;
end;

//****************************************************
function TNotesAgent.GetRunOnWeekends: boolean;
begin
  result := ASSISTODS_FLAG_NOWEEKENDS = (FAgentInfo.dwFlags and integer(ASSISTODS_FLAG_NOWEEKENDS));
end;

//****************************************************
procedure TNotesAgent.SetRunOnWeekends(Value: boolean);
begin
  if (Value = true) then
    FAgentInfo.dwFlags := FAgentInfo.dwFlags xor ASSISTODS_FLAG_NOWEEKENDS
  else
    FAgentInfo.dwFlags := FAgentInfo.dwFlags or ASSISTODS_FLAG_NOWEEKENDS;
end;

//****************************************************
function TNotesAgent.GetChooseWhenEnabled: boolean;
begin
  result := ASSISTODS_FLAG_CHOOSEWHENENABLED = (FAgentInfo.dwFlags and integer(ASSISTODS_FLAG_CHOOSEWHENENABLED));
end;

//****************************************************
procedure TNotesAgent.SetChooseWhenEnabled(Value: boolean);
begin
        if (Value = true) then
                FAgentInfo.dwFlags := FAgentInfo.dwFlags xor ASSISTODS_FLAG_CHOOSEWHENENABLED
        else
                FAgentInfo.dwFlags := FAgentInfo.dwFlags or ASSISTODS_FLAG_CHOOSEWHENENABLED;
end;

//****************************************************
function TNotesAgent.GetStoreHighLights: boolean;
begin
       result := ASSISTODS_FLAG_STOREHIGHLIGHTS = (FAgentInfo.dwFlags and integer(ASSISTODS_FLAG_STOREHIGHLIGHTS));
end;
//****************************************************
procedure TNotesAgent.SetStoreHighLights(Value: boolean);
begin
        if (Value = true) then
                FAgentInfo.dwFlags := FAgentInfo.dwFlags xor ASSISTODS_FLAG_STOREHIGHLIGHTS
        else
                FAgentInfo.dwFlags := FAgentInfo.dwFlags or ASSISTODS_FLAG_STOREHIGHLIGHTS;
end;

//****************************************************
function TNotesAgent.GetAssistFlags: string;
begin
  Result := Items[ASSIST_FLAGS].AsString;
end;

//****************************************************
procedure TNotesAgent.SetAssistFlags(const Value: string);
begin
  Items[ASSIST_FLAGS].IsSigned := true;
  Items[ASSIST_FLAGS].AsString := Value;

  // Also, if the flag is ASSIST_FLAG_ENABLED then we also
  // need to set the dwFlags item of ods_assiststruct (sick....)
  //SetIsHidden( True );
end;

//****************************************************
function TNotesAgent.GetAssistFlag(Flag: string): boolean;
begin
  Result := pos(Flag, GetAssistFlags) <> 0;
end;

//****************************************************
procedure TNotesAgent.SetAssistFlag(Flag: string; Value: boolean);
var
  n: integer;
  s: string;
begin
  s := GetAssistFlags;
  if Value and (pos(Flag, s) = 0) then SetAssistFlags(s + Flag)
  else begin
    n := pos(Flag, s);
    if n > 0 then delete(s,n,1);
    SetAssistFlags( s );
  end;
end;

//****************************************************
procedure TNotesAgent.SetAgentType;
var
    TempNumber: NUMBER;
begin
        TempNumber := wActionType;
        // Remove old item (if any)
        NSFItemDelete(Handle, ASSIST_TYPE_ITEM, Length(ASSIST_TYPE_ITEM));

        CheckError(NSFItemAppend(
                Handle,                        // handle to note to append to */
                ITEM_SIGN Or ITEM_SUMMARY,      // item flags */
                ASSIST_TYPE_ITEM,              // item name: "$AssistType" */
                length(ASSIST_TYPE_ITEM),
                TYPE_NUMBER,                   // item type */
                @TempNumber,                   // item value */
                sizeof(TempNumber)));   // value length */
end;

procedure TNotesAgent.SetAgentVersion;
var
    CurrTimeDate:TIMEDATE;
begin
        OSCurrentTIMEDATE(@CurrTimeDate);
        // Remove old item (if any)
        NSFItemDelete(Handle, ASSIST_VERSION_ITEM, Length(ASSIST_VERSION_ITEM));

        NSFItemSetTime(Handle,
                      ASSIST_VERSION_ITEM,
                      @CurrTimeDate);

end;
//****************************************************
procedure TNotesAgent.SetAgentLastInfo;
var
    EmptyTimeDate:TIMEDATE;
    DocCount:NUMBER;
begin
    DocCount := 0;

   // set empty (Initial) Last Run info
    EmptyTimeDate.T1 := 0;
    EmptyTimeDate.T2 := 0;

// Remove old item (if any)
   NSFItemDelete(Handle, ASSIST_LASTRUN_ITEM, Length(ASSIST_LASTRUN_ITEM));

    CheckError(NSFItemAppend(
                Handle,                             // handle to note to append to */
                ITEM_SIGN Or ITEM_SUMMARY,           // item flags */
                ASSIST_LASTRUN_ITEM,                // $AssistLastRun */
                length(ASSIST_LASTRUN_ITEM),
                TYPE_TIME,                          // item type */
                @EmptyTimeDate,                     // item value */
                sizeof(EmptyTimeDate)));            // value length */

    // set empty DocCountRun info
    // Remove old item (if any)
   NSFItemDelete(Handle, ASSIST_DOCCOUNT_ITEM, Length(ASSIST_DOCCOUNT_ITEM));

    CheckError(NSFItemAppend(
                Handle,                             // handle to note to append to */
                ITEM_SIGN Or ITEM_SUMMARY,           // item flags */
                ASSIST_DOCCOUNT_ITEM,               // "$AssistDocCount" */
                length(ASSIST_DOCCOUNT_ITEM),
                TYPE_NUMBER,                        // item type */
                @DocCount,                          // item value */
                sizeof(DocCount)));         // value length */   
end;

//****************************************************
procedure TNotesAgent.SetAgentTrigger;
begin
  // Remove old item (if any)
  NSFItemDelete(Handle, ASSIST_TRIGGER_ITEM, Length(ASSIST_TRIGGER_ITEM));
  CheckError(NSFItemSetText(
              Handle,
              ASSIST_TRIGGER_ITEM,   // "$AssistTrigger" */
              pchar(inttostr(wTrigger)),
              1));
end;


//****************************************************
procedure TNotesAgent.SetAgentAction;
var
    AgentQuery: CDQUERYHEADER;
begin


// First append "empty" TYPE_QUERY item since we assume no front-end query
    AgentQuery.Header.Signature := SIG_QUERY_HEADER;
    AgentQuery.Header.Length := ODSLength(_CDQUERYHEADER);
    AgentQuery.dwFlags := 0;

// Remove old item (if any)
    NSFItemDelete(Handle, ASSIST_QUERY_ITEM, Length(ASSIST_QUERY_ITEM));

    CheckError(NSFItemAppend(
                Handle,                            // handle to note to append to
                ITEM_SIGN,                         // item flags
                ASSIST_QUERY_ITEM,                 // item name: "$AssistQuery"
                length(ASSIST_QUERY_ITEM),
                TYPE_QUERY,                        // item type
                @AgentQuery,                       // item value
                AgentQuery.Header.Length));// value length


// Then apppend TYPE_ACTION item
// Remove old item (if any)
   NSFItemDelete(Handle, ASSIST_ACTION_ITEM, Length(ASSIST_ACTION_ITEM));

    CheckError(NSFItemAppend(
                Handle,                           // handle to note to append to */
                ITEM_SIGN,                        // item flags */
                ASSIST_ACTION_ITEM,               //* item name: "$AssistAction" */
                length(ASSIST_ACTION_ITEM),
                TYPE_ACTION,                      // item type */
                AgentAction,                      // item value */
                dwActionLen ));                  //* value length */

// Lastly, append "empty" TYPE_LSOBJECT (extra Action data) item
// Remove old item (if any)
   NSFItemDelete(Handle, ASSIST_EXACTION_ITEM, Length(ASSIST_EXACTION_ITEM));
    CheckError(NSFItemAppend(
                Handle,                           // handle to note to append to
                ITEM_SIGN,                       // item flags
                ASSIST_EXACTION_ITEM,            // item name: "$AssistAction_Ex" */
                length(ASSIST_EXACTION_ITEM),
                TYPE_LSOBJECT,                   // item type */
                AgentAction,                     // item value */
                sizeof(WORD)));                   // value length */

end;

//****************************************************
procedure TNotesAgent.GetAgentInfoStruct;
var
  dwItemSize: dword;
  wDataType: word;
  bval: BLOCKID;
  pBuf: pchar;
begin
  CheckError(NSFItemInfo(Handle, ASSIST_INFO_ITEM,
              length (ASSIST_INFO_ITEM), nil,
              @wDataType, @bval, @dwItemSize));

  pBuf := OsLockBlock(bval);
  try
    // Read in the assiststruct ODS record and ensure that there are data objects
    pBuf := pbuf + sizeOf(word);
    ODSReadMemory( @pBuf, _ODS_ASSISTSTRUCT, @FAgentInfo, 1 );
  finally
    OSUnlockBlock(bval);
  end;
end;

//****************************************************
procedure TNotesAgent.SetAgentInfoStruct;
var
  pBuffer: Pointer;
  hItem: longint;
  dwItemLen: DWORD;
  bhValue:BLOCKID;
begin
  // Allocate an object to hold the ods object and its TYPE
  dwItemLen := DWord( sizeof(WORD) + ODSLength(_ODS_ASSISTSTRUCT) );
  CheckError(OSMemAlloc(MEM_GROWABLE, dwItemLen, @hItem));
  try
    pBuffer := OSLockObject(LHandle(hItem));

    // Set the TYPE word in HOST format
    PWord(pBuffer)^ :=   Word(TYPE_ASSISTANT_INFO);
    pBuffer := pword (DWord(pBuffer) + sizeOf(word));

    ODSWriteMemory(@pBuffer, _ODS_ASSISTSTRUCT ,@FAgentInfo, 1);

    OSUnlockObject(LHandle(hItem));

    // Create a BLOCKID and fill it
    bhValue.pool := hItem;
    bhValue.block := Word(0);

    // Remove old item (if any)
    NSFItemDelete(Handle, ASSIST_INFO_ITEM, Length(ASSIST_INFO_ITEM));

    // Save the item to the note
    CheckError(NSFItemAppendByBLOCKID(Handle,  // handle to note to append
                    ITEM_SIGN,                 // item flags
                    ASSIST_INFO_ITEM,         // item name: "$AssistInfo"
                    length(ASSIST_INFO_ITEM),  // item length
                    bhValue,                   // BLOCKID with value
                    dwItemLen, nil));        // value length
    except
      OSMemFree(hItem);
      raise;
    end;
end;

//****************************************************
procedure TNotesAgent.SetAgentRunInfo;
var
  isPrivate: boolean;
  hRunInfo: LHandle;
  dwObjectSize,dwObjectID,dwItemSize: DWord;
  pw, pRunInfo: Pointer;
  bidRunInfo:BLOCKID;
  objRunInfo:OBJECT_DESCRIPTOR;
  RunHeader:ODS_ASSISTRUNOBJECTHEADER;
  RunEntry:ODS_ASSISTRUNOBJECTENTRY;
begin
  isPrivate := GetIsPrivate;

  dwObjectSize := ODSLength(_ODS_ASSISTRUNOBJECTHEADER) +
                  ODSLength(_ODS_ASSISTRUNOBJECTENTRY);

  // if Private agent, then allocate private, filter note
  if isPrivate then begin
    CheckError(NSFDbAllocObject(Database.Handle, dwObjectSize,
                                 (NOTE_CLASS_FILTER Or NOTE_CLASS_PRIVATE),
                                 0, dwObjectID));
  end
  else begin
    // Shared agent use public data note
    CheckError(NSFDbAllocObject(Database.Handle, dwObjectSize,
                                 NOTE_CLASS_DOCUMENT,
                                 0, dwObjectID));
  end;
  try
    // set up OBJECT_DESCRIPTOR info
    FillChar(objRunInfo,sizeOf(OBJECT_DESCRIPTOR), 0);
    objRunInfo.ObjectType := OBJECT_ASSIST_RUNDATA;
    objRunInfo.RRV := dwObjectID;

    dwItemSize := DWord( ODSLength(_WORD) + ODSLength(_OBJECT_DESCRIPTOR) );

    CheckError(OSMemAlloc(0, dwItemSize, @(bidRunInfo.pool)));
    try
      bidRunInfo.block := 0;

      // Set RunInfo OBJECT_TYPE to TYPE_OBJECT
      pRunInfo := OSLockBlock(bidRunInfo);

      // Set the TYPE word in HOST format
      pword(pRunInfo)^ := word(TYPE_OBJECT);
      pRunInfo := pointer (dword(pRunInfo) + sizeOf(word));

      // and copy the OBJECT_DESCRIPTOR
      ODSWriteMemory (@pRunInfo, _OBJECT_DESCRIPTOR, @objRunInfo, 1);
      OSUnlockBlock(bidRunInfo);

      // Remove old item (if any)
      NSFItemDelete(Handle, ASSIST_RUNINFO_ITEM, Length(ASSIST_RUNINFO_ITEM));

      // finally append RunInfo object item
      CheckError(NSFItemAppendObject(Handle, ITEM_SUMMARY,
                              ASSIST_RUNINFO_ITEM,
                              length(ASSIST_RUNINFO_ITEM),
                              bidRunInfo,
                              dwItemSize,
                              TRUE));      // Domino and Notes will deallocate memory */
    except
      OSMemFree(bidRunInfo.pool);
      raise;
    end;

    //set up initial Run object header and entry placeholder
    CheckError(OSMemAlloc(0, dwObjectSize, @hRunInfo));
    try
      pw := OSLockObject(hRunInfo);

      FillChar(RunHeader, sizeof(ODS_ASSISTRUNOBJECTHEADER), 0);
      RunHeader.wEntries := Word(1);  // placeholder entry
      ODSWriteMemory(@pw, _ODS_ASSISTRUNOBJECTHEADER, @RunHeader, 1);

      FillChar(RunEntry, sizeof(ODS_ASSISTRUNOBJECTENTRY), 0);
      ODSWriteMemory(@pw, _ODS_ASSISTRUNOBJECTENTRY, @RunEntry, 1);
      OSUnlockObject(hRunInfo);

      // and add to object
      CheckError(NSFDbWriteObject(Database.Handle, dwObjectID, hRunInfo, 0, dwObjectSize));
    finally
      OSMemFree(hRunInfo);
    end;
  except
    NSFDbFreeObject(Database.Handle, dwObjectID);
    raise;
  end;
end;

//****************************************************
procedure TNotesAgent.SetType(const Value: TSourceCodeType);
begin
  // Always set this to be a Version 4 and higher agent
  // (don't care about V3 or less)
  inherited;
  SetFlag(DESIGN_FLAG_HIDE_FROM_V3, true);
  SetFlag(DESIGN_FLAG_V4AGENT, true);
  SetFlag(DESIGN_FLAG_LOTUSSCRIPT_AGENT, false);
  SetFlag(DESIGN_FLAG_JAVA_AGENT_WITH_SOURCE, false);
  SetFlag(DESIGN_FLAG_JAVA_AGENT, false);

  case Value of
    scFormula: begin
      SetAgentType(SIG_ACTION_FORMULA);
    end;

    scLotusScript: begin
      SetFlag(DESIGN_FLAG_LOTUSSCRIPT_AGENT, true);
      SetAgentType(SIG_ACTION_LOTUSSCRIPT);
    end;

    scJava: begin
      SetFlag(DESIGN_FLAG_JAVA_AGENT_WITH_SOURCE, true);
      SetAgentType(SIG_ACTION_JAVAAGENT);
    end;

    scCompiledJava: begin
      SetFlag(DESIGN_FLAG_JAVA_AGENT, true);
      SetAgentType(SIG_ACTION_JAVAAGENT);
    end;
  end;
end;


//****************************************************
procedure TNotesAgent.buildActionLS;
var
  AgentAction, buff_ptr: PByte;
  dwBuffLen:DWORD;
begin
  GetMem(AgentAction,  ODSLength (_CDACTIONHEADER) +
                           ODSLength (_CDACTIONLOTUSSCRIPT) +
                           strlen(pchar(FAgentSourceCode)) + 1);
  try
    buff_ptr := AgentAction;
    FAgentHeader.Header.Signature := SIG_ACTION_HEADER;
    FAgentHeader.Header.Length := ODSLength(_CDACTIONHEADER);
    ODSWriteMemory( @buff_ptr, _CDACTIONHEADER, @FAgentHeader, 1 );

    FAgentScript.Header.Signature := SIG_ACTION_LOTUSSCRIPT;
    FAgentScript.Header.Length := ODSLength(_CDACTIONLOTUSSCRIPT) +
                                  strlen(PChar(FAgentSourceCode)) + 1;
    FAgentScript.dwFlags := DWORD( 0 );
    FAgentScript.dwScriptLen := DWORD(strlen(PChar(FAgentSourceCode)) + 1);

    ODSWriteMemory( @buff_ptr, _CDACTIONLOTUSSCRIPT, @FAgentScript, 1 );
    move(pchar(FAgentSourceCode)^, buff_ptr^, FAgentScript.dwScriptLen);
    buff_ptr := pointer(dword(buff_ptr) + FAgentScript.dwScriptLen) ;
    dwBuffLen := DWORD(   PChar(dword( buff_ptr )) - PChar(AgentAction) );

    SetAgentAction(AgentAction, dwBuffLen);
  finally
    FreeMem (AgentAction);
  end;
end;

//****************************************************

procedure TNotesAgent.buildActionFormula;
var
        AgentAction, buff_ptr: PByte;
        dwBuffLen:DWORD;
begin
  // Just to make sure, compile the formula. This is very important.
  // If the formula is not compiled before this point, it will go
  // boom.
   CompileFormula(FAgentSourceCode);

   GetMem(AgentAction,  ODSLength (_CDACTIONHEADER) +
                             ODSLength (_CDACTIONFORMULA) +
                             FFormulaLength);
   try
     buff_ptr := AgentAction;

     FAgentHeader.Header.Signature := SIG_ACTION_HEADER;
     FAgentHeader.Header.Length := ODSLength(_CDACTIONHEADER);
     ODSWriteMemory( @buff_ptr, _CDACTIONHEADER, @FAgentHeader, 1 );

     FAgentFormula.Header.Signature := SIG_ACTION_FORMULA;
     FAgentFormula.Header.Length := ODSLength(_CDACTIONFORMULA) +
                                    FFormulaLength;
     FAgentFormula.dwFlags := DWORD( 0 );
     FAgentFormula.wFormulaLen := FFormulaLength;

     ODSWriteMemory( @buff_ptr, _CDACTIONFORMULA, @FAgentFormula, 1 );
     move(OSLockObject(FFormulaHandle)^, buff_ptr^, FAgentFormula.wFormulaLen);
     buff_ptr := pointer(dword(buff_ptr) + FAgentFormula.wFormulaLen) ;
     dwBuffLen := DWORD(   PChar(dword( buff_ptr )) - PChar(AgentAction) );

     SetAgentAction(AgentAction, dwBuffLen);

     // Dont forget to free agent formula handle here
     OSUnlockObject(FFormulaHandle);
     OSMemFree(FFormulaHandle);
  finally
   FreeMem (AgentAction);
  end;
end;

//****************************************************
procedure TNotesAgent.buildActionJava;
var
        idx: Integer;
        AgentAction, buff_ptr: PByte;
        dwBuffLen:DWORD;
begin
  GetMem(AgentAction,  ODSLength (_CDACTIONHEADER) +
                           ODSLength (_CDACTIONJAVAAGENT) +
                           Length(FJavamainClassName) +
                           Length(FJavaCodePath) +
                           Length(FJavaClassNameList) + 2);
  try
    buff_ptr := AgentAction;

    FAgentHeader.Header.Signature := SIG_ACTION_HEADER;
    FAgentHeader.Header.Length := ODSLength(_CDACTIONHEADER);
    ODSWriteMemory( @buff_ptr, _CDACTIONHEADER, @FAgentHeader, 1 );

    FAgentJava.Header.Signature := SIG_ACTION_JAVAAGENT;
    FAgentJava.Header.Length := ODSLength(_CDACTIONJAVAAGENT) +
                                   Length(FJavamainClassName) +
                                   Length(FJavaCodePath) +
                                   Length(FJavaClassNameList) + 2;

    FAgentJava.wClassNameLen := length(FJavamainClassName);
    FAgentJava.wCodePathLen := length(FJavaCodePath);
    FAgentJava.wFileListBytes := length(FJavaClassNameList) + 2;
    FAgentJava.wLibraryListBytes := 0; // ??

    ODSWriteMemory( @buff_ptr, _CDACTIONJAVAAGENT, @FAgentJava, 1 );
    move(PChar(FJavamainClassName)^, pchar(buff_ptr)^, FAgentJava.wClassNameLen);
    buff_ptr := pointer(dword(buff_ptr) + FAgentJava.wClassNameLen);

    move(PChar(FJavaCodePath)^, buff_ptr^, FAgentJava.wCodePathLen);
    buff_ptr := pointer(dword(buff_ptr) + FAgentJava.wCodePathLen);

    move(PChar(FJavaClassNameList)^, buff_ptr^, FAgentJava.wFileListBytes);
    buff_ptr := pointer(dword(buff_ptr) + FAgentJava.wFileListBytes);

    dwBuffLen := DWORD(   PChar(dword( buff_ptr )) - PChar(AgentAction) );

    SetAgentAction(AgentAction, dwBuffLen);
  finally
    FreeMem (AgentAction);
  end;

  if (FJavaCodePath <> '') and (FJavaCodePath[length(FJavaCodePath) - 1] <> '\')
    then FJavaCodePath := FJavaCodePath + '\';

  // Attach the class file(s)
  for idx := 0 to System.High(FJavaClassNameArray) do
    CheckError( NSFNoteAttachFile(Handle,
               ITEM_NAME_ATTACHMENT,
               WORD(length(ITEM_NAME_ATTACHMENT)),
               PChar(FJavaCodePath + FJavaClassNameArray[idx]),
               PChar(FJavaClassNameArray[idx]),
               HOST_LOCAL or COMPRESS_NONE) );
end;
//****************************************************

procedure TNotesAgent.Save(force, createResponse, markRead: boolean);
begin
  SetAgentInfoStruct;
  
  // $AssistRunInfo
  SetAgentRunInfo;

  // Signature is required
  Sign;

  // Call up the chain...
  inherited Save(force, createResponse, markRead);
end;

procedure TNotesAgent.SetDayOfMonthOrWeek(const Value: word);
begin
  //
end;

procedure TNotesAgent.SetEndDate(const Value: TDateTime);
begin
  FAgentInfo.EndTime := DateTimeToNotes( endDate );
end;

procedure TNotesAgent.SetStartDate(const Value: TDateTime);
begin
  FAgentInfo.StartTime := DateTimeToNotes( Value );
end;

function TNotesAgent.GetEndTime: TDateTime;
var
  h, m, s, ms: word;
begin
  ms := 0;
  h := FAgentInfo.dwTime2 div TICKS_IN_HOUR;
  m := (FAgentInfo.dwTime2 - (h * TICKS_IN_HOUR)) div TICKS_IN_MINUTE;
  s := (FAgentInfo.dwTime2 - (h * TICKS_IN_HOUR) - (m * TICKS_IN_MINUTE)) div TICKS_IN_SECOND;
  Result := EncodeTime(h, m, s, ms);
end;

function TNotesAgent.GetStartTime: TDateTime;
var
  h, m, s, ms: word;
begin
  ms := 0;
  h := FAgentInfo.dwTime1 div TICKS_IN_HOUR;
  m := (FAgentInfo.dwTime1 - (h * TICKS_IN_HOUR)) div TICKS_IN_MINUTE;
  s := (FAgentInfo.dwTime1 - (h * TICKS_IN_HOUR) - (m * TICKS_IN_MINUTE)) div TICKS_IN_SECOND;
  Result := EncodeTime(h, m, s, ms);
end;

procedure TNotesAgent.SetEndTime(const Value: TDateTime);
var
  h, m, s, ms: word;
begin
  DecodeTime(Value, h, m, s, ms);
  FAgentInfo.dwTime2 := h*TICKS_IN_HOUR +
                        m*TICKS_IN_MINUTE +
                        s*TICKS_IN_SECOND;
end;

procedure TNotesAgent.SetStartTime(const Value: TDateTime);
var
  h, m, s, ms: word;
begin
  DecodeTime(Value, h, m, s, ms);
  FAgentInfo.dwTime1 := h*TICKS_IN_HOUR +
                        m*TICKS_IN_MINUTE;
end;

//****************************************************
// Database Script Class Added by Daniel 2002-11-26
//****************************************************
const
  DBLIB_CODE_ITEM = '$DBScript';
  
constructor TNotesDBScript.CreateNew(notesDatabase: TNotesDatabase);
var
  wNoteClass:word;
begin
  inherited CreateNew(notesDatabase);
  // global DB script is also a filter like an agent...
    wNoteClass := (NOTE_CLASS_FILTER);
    NSFNoteSetInfo(Handle, _NOTE_CLASS, @wNoteClass);
    Items['$PublicAccess'].AsString := '1';
    SetDesignFlags('t34Q'); //= DatabaseScript HideFromV3 HideFromV4 QueryFilter
    Title := 'Database Script'; // Should always have this name
end;

//****************************************************

function  CodeScanProcDB(Spare, ItemFlags: word;
                              Name: PChar;
                              NameLength: word;
                              Value: pointer;
                              ValueLength: dword;
                              RoutineParameter: pointer): STATUS;  stdcall;
var
  s, v: string;
begin
  s := Lmbcs2Native(Copy(strPas(name), 1, NameLength));
  if compareText(s, DBLIB_CODE_ITEM) = 0 then begin
    v := NotesToString(pchar(DWORD(Value)+sizeof(WORD)),ValueLength-sizeof(WORD));
    TStringList(RoutineParameter).Add(v);
  end;
  Result := NOERROR;
end;

function TNotesDBScript.GetSourceCode: string;
var
  s: TStringList;
begin
  s := TStringList.Create;
  try
    CheckError(NsfItemScan(Handle, CodeScanProcDB, s));
    Result := s.text;
  finally
    s.free;
  end;
end;

//****************************************************
procedure TNotesDBScript.SetSourceCode(Value: string);
var
        formattedVal: String;
begin
// Convert the code to a format compliant with the Notes IDE
  formattedVal := FormatLSCode(PChar(Native2Lmbcs(Value)) );

  // Remove old item (if any)
   NSFItemDelete(Handle, DBLIB_CODE_ITEM, Length(DBLIB_CODE_ITEM));

  Items[DBLIB_CODE_ITEM].IsSigned := true;
  Items[DBLIB_CODE_ITEM].IsSummary := false;
  Items[DBLIB_CODE_ITEM].AsPChar:= PChar(formattedVal);
end;

end.
