{==============================================================================|
| Project : Notes/Delphi class library                           | 3.9         |
|==============================================================================|
| Content:                                                                     |
|==============================================================================|
| The contents of this file are subject to the Mozilla Public License Ver. 1.0 |
| (the "License"); you may not use this file except in compliance with the     |
| License. You may obtain a copy of the License at http://www.mozilla.org/MPL/ |
|                                                                              |
| Software distributed under the License is distributed on an "AS IS" basis,   |
| WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for |
| the specific language governing rights and limitations under the License.    |
|==============================================================================|
| Initial Developers of the Original Code are:                                 |
|   Winalot                                                                    |
| All Rights Reserved.                                                         |
|   Last Modified:                                                             |
|     29.11.00                                                                 |
|==============================================================================|
| Contributors and Bug Corrections:                                            |
|   Sergey Kolchin (skolchin@yahoo.com)                                        |
|==============================================================================|
| History: see README.TXT                                                      |
|==============================================================================|
| Notes UI add-ins support unit                                                |
| See Notes C API Reference/User Guide from Lotus for description of UI add-in |
| functionality                                                                |
|                                                                              |
| WARNING! Add-in DLLs which use N/D library must be compiled with             |
| NO_INIT_SECTION define symbol. Otherwise they could crash the Notes client   |
| upon exit.                                                                   |
|                                                                              |
| Latest versions is available at                                              |
|   http://www.geocities.com/skolchin/                                         |
|==============================================================================|}
unit Util_AddinMenu;

interface

Uses Windows, Messages, Util_Lnapi;

const
  MAX_ADDIN_MENU_NAME = 64;
  OLDMAXPATH = 100;
  MAXHEADERSTRING = 256;
  MAXFACESIZE = 32;
  UNTITLED_SIZE = 30;
  FULLNAME_SIZE = 255;

  // Notes AddIn Menu operations
  NAMM_INIT	=	0;
  NAMM_INITMENU	=	1;
  NAMM_COMMAND	=	2;
  NAMM_TERM	=	3;

  // Notes AddIn Menu return values
  NAM_NOERROR	=	0;
  NAM_INIT_CONTINUE=	1;
  NAM_INIT_STOP	=	2;

  //  Notes AddIn menu messages
  WM_ADDIN_FIRST = (WM_USER+400);
  WM_ADDIN_REFRESH = (WM_ADDIN_FIRST+0);
  WM_ADDIN_IMPORT = (WM_ADDIN_FIRST+1);

  // Notes context flags
  NAM_IN_DESK	=	0;
  NAM_IN_VIEW	=	1;
  NAM_IN_VIEW_DESIGN=	2;
  NAM_IN_EDIT_RO	=	3;
  NAM_IN_EDIT_RW	=	4;
  NAM_IN_EDIT_DESIGN =	5;

type
  NAMRESULT = WORD;
  HMENU = HWND;
  Fontid = dword;
  FLAG = WORD;

  // This file contains the data structures for the info passed from
  // the view setup module to import/export exe's
  VIEWIXDATA = packed record
    hFormNames : Handle;
    FormCount : WORD;
    Reserved1 : WORD;
    hNoteFile : DbHandle;
    hViewFile : DbHandle;
    wFlags : WORD;
    ViewNoteID : NOTEID;
    ViewNoteUNID : UNID;
    hViewCollection : HCOLLECTION;
    Reserved3 : WORD;
    hUnreadList : HANDLE;
    hCollapsedList : HANDLE;
    hSelectedList : HANDLE;
    Reserved4 : WORD;
    ViewIsPrivate : FLAG;
    CaretNoteID : NOTEID;
    ColumnCount : WORD;
    hColDesc : HANDLE;
    LineArrayEntrySize : WORD;
    TextWidth : WORD;
    NumTextWidth : WORD;
    Categories : WORD;
    hFormNoteIDs : HANDLE;
    hFormNamesFull : HANDLE;
  end;
  PVIEWIXDATA = ^VIEWIXDATA;

  EDITIMPORTDATA = TEDITIMPORTDATA;                  //See Util_LnApi
  PEDITIMPORTDATA = ^EDITIMPORTDATA;

  HEAD_DESC = packed record
    FontPitchAndFamily: BYTE;
    FontName: Array[0..MAXFACESIZE-1] of Char;
    Font: FONTID;
    HeadLength: WORD;
  end;
  PHEAD_DESC = ^HEAD_DESC;

  HEAD_DESC_BUFFER = packed record
    Desc: HEAD_DESC;
    aString: Array[0..MAXHEADERSTRING-1] of Char; //* Must be terminated by '\0' */
  end;
  PHEAD_DESC_BUFFER = ^HEAD_DESC_BUFFER;

  PPRINT_SETTINGS = ^PRINT_SETTINGS;

  EDITEXPORTDATA = TEDITEXPORTDATA;
  PEDITEXPORTDATA = ^EDITEXPORTDATA;

  EDITIXDATA = packed record
    Import : EDITIMPORTDATA;
    aExport : EDITEXPORTDATA;
    CaretFieldName: Array[0..34-1] of Char;
    CaretFieldType: WORD;
  end;
  PEDITIXDATA = ^EDITIXDATA;

  NAM_CONTEXT_DATA = packed record
    //This field contains the following bit fields of original structure:
    //  Context:3
    //  fCanExport:1;
    //  fCanImport:1;
    // Use GetContext, GetCanExport, GetCanImport functions to retrieve bits
    Context: DWORD;
    case boolean of
      True:(Edit:EDITIXDATA);
      False:(View:VIEWIXDATA);
  end;
  PNAM_CONTEXT_DATA = ^NAM_CONTEXT_DATA;

  NAM_INIT_INFO = packed record
    wStartingID: WORD;                      //* Input: DLL must add this to its ID. */
    wMenuItemNumber: WORD;        //* Input: Ascending ordinal number for each call. */
    wMenuID: WORD;                //* Output: DLL's Menu ID. */
    MenuItemName: Array[0..MAX_ADDIN_MENU_NAME-1] of Char;                //* Output: Text of menu item. */
  end;
  PNAM_INIT_INFO = ^NAM_INIT_INFO;

  NAMPROC = function (Command : WORD; pParam : Pointer): STATUS; stdcall;

  NAM_COMMAND_INFO = packed record
    hNotesWnd: HWND;                      //* Handle of Notes main window to use as parent. */
    wMenuID: WORD;        //* DLL's ID number of selected menu item. */
    Data: NAM_CONTEXT_DATA;                //* Context specific data */
  end;
  PNAM_COMMAND_INFO = ^NAM_COMMAND_INFO;

  NAM_INITMENU_INFO = packed record
    hMenu: HMENU;                      //* Handle of the Add-In PopUp menu. */
    Data: NAM_CONTEXT_DATA;        //*	Context specific data. */
  end;
  PNAM_INITMENU_INFO = ^NAM_INITMENU_INFO;

function GetContext(const data: NAM_CONTEXT_DATA): word;
function GetCanImport(const data: NAM_CONTEXT_DATA): boolean;
function GetCanExport(const data: NAM_CONTEXT_DATA): boolean;

implementation

function GetContext(const data: NAM_CONTEXT_DATA): word;
begin
  Result := (data.Context and 7);     //binary 000111
end;

function GetCanImport(const data: NAM_CONTEXT_DATA): boolean;
begin
  Result := ((data.Context shr 3) and 1) <> 0;
end;

function GetCanExport(const data: NAM_CONTEXT_DATA): boolean;
begin
  Result := ((data.Context shr 4) and 1) <> 0;
end;


end.
